@@ -17,13 +17,8 @@
 //
 //===----------------------------------------------------------------------===//
 
-#include <cstdio>
-#include <cstring>
-#include <iomanip>
-#include <iostream>
 #include <map>
 #include <set>
-#include <sstream>
 #include <string>
 
 #include "llvm-version.h"
@@ -51,21 +46,24 @@
 #include <llvm/Support/MemoryBuffer.h>
 #include <llvm/Support/SourceMgr.h>
 #include <llvm/Support/TargetRegistry.h>
-#include "llvm/Support/TargetSelect.h"
+#include <llvm/Support/TargetSelect.h>
 #include <llvm/Support/raw_ostream.h>
-#include "llvm/Support/FormattedStream.h"
+#include <llvm/Support/FormattedStream.h>
+#include <llvm/Support/NativeFormatting.h>
 #include <llvm/ExecutionEngine/JITEventListener.h>
 #include <llvm/IR/LLVMContext.h>
 #include <llvm/DebugInfo/DIContext.h>
-#include "llvm/DebugInfo/DWARF/DWARFContext.h"
+#include <llvm/DebugInfo/DWARF/DWARFContext.h>
 #include <llvm/IR/DebugInfo.h>
 #include <llvm/IR/Function.h>
 #include <llvm/IR/Module.h>
 #include <llvm/IR/IntrinsicInst.h>
-#include "llvm/IR/AssemblyAnnotationWriter.h"
+#include <llvm/IR/AssemblyAnnotationWriter.h>
+#include <llvm/IR/LegacyPassManager.h>
 
 #include "julia.h"
 #include "julia_internal.h"
+#include "jitlayers.h"
 #include "processor.h"
 
 using namespace llvm;
@@ -168,7 +166,6 @@ void DILineInfoPrinter::emit_lineinfo(raw_ostream &Out, std::vector<DILineInfo>
 {
     if (verbosity == output_none)
         return;
-    bool update_line_only = false;
     uint32_t nframes = DI.size();
     if (nframes == 0)
         return; // just skip over lines with no debug info at all
@@ -181,21 +178,36 @@ void DILineInfoPrinter::emit_lineinfo(raw_ostream &Out, std::vector<DILineInfo>
             break;
         }
     }
-    if (collapse_recursive && 0 < nctx) {
-        // check if we're adding more frames with the same method name,
-        // if so, drop all existing calls to it from the top of the context
-        // AND check if instead the context was previously printed that way
-        // but now has removed the recursive frames
-        StringRef method = StringRef(context.at(nctx - 1).FunctionName).rtrim(';');
-        if ((nctx < nframes && StringRef(DI.at(nframes - nctx - 1).FunctionName).rtrim(';') == method) ||
-            (nctx < context.size() && StringRef(context.at(nctx).FunctionName).rtrim(';') == method)) {
-            update_line_only = true;
-            while (nctx > 0 && StringRef(context.at(nctx - 1).FunctionName).rtrim(';') == method) {
-                nctx -= 1;
+    bool update_line_only = false;
+    if (collapse_recursive) {
+        if (nctx > 0) {
+            // check if we're adding more frames with the same method name,
+            // if so, drop all existing calls to it from the top of the context
+            // AND check if instead the context was previously printed that way
+            // but now has removed the recursive frames
+            StringRef method = StringRef(context.at(nctx - 1).FunctionName).rtrim(';');
+            if ((nctx < nframes && StringRef(DI.at(nframes - nctx - 1).FunctionName).rtrim(';') == method) ||
+                (nctx < context.size() && StringRef(context.at(nctx).FunctionName).rtrim(';') == method)) {
+                update_line_only = true;
+                while (nctx > 0 && StringRef(context.at(nctx - 1).FunctionName).rtrim(';') == method) {
+                    nctx -= 1;
+                }
             }
         }
+        else if (context.size() > 0) {
+            update_line_only = true;
+        }
     }
-    // examine what frames we're returning from
+    else if (nctx < context.size() && nctx < nframes) {
+        // look at the first non-matching element to see if we are only changing the line number
+        const DILineInfo &CtxLine = context.at(nctx);
+        const DILineInfo &FrameLine = DI.at(nframes - 1 - nctx);
+        if (CtxLine.FileName == FrameLine.FileName &&
+                StringRef(CtxLine.FunctionName).rtrim(';') == StringRef(FrameLine.FunctionName).rtrim(';')) {
+            update_line_only = true;
+        }
+    }
+    // examine how many frames we're returning from
     if (nctx < context.size()) {
         // compute the new inlining depth
         uint32_t npops;
@@ -212,15 +224,6 @@ void DILineInfoPrinter::emit_lineinfo(raw_ostream &Out, std::vector<DILineInfo>
         else {
             npops = context.size() - nctx;
         }
-        // look at the first non-matching element to see if we are only changing the line number
-        if (!update_line_only && nctx < nframes) {
-            const DILineInfo &CtxLine = context.at(nctx);
-            const DILineInfo &FrameLine = DI.at(nframes - 1 - nctx);
-            if (CtxLine.FileName == FrameLine.FileName &&
-                    StringRef(CtxLine.FunctionName).rtrim(';') == StringRef(FrameLine.FunctionName).rtrim(';')) {
-                update_line_only = true;
-            }
-        }
         context.resize(nctx);
         update_line_only && (npops -= 1);
         if (npops > 0) {
@@ -228,46 +231,26 @@ void DILineInfoPrinter::emit_lineinfo(raw_ostream &Out, std::vector<DILineInfo>
             Out << LineStart << inlining_indent("│") << repeat{npops, "└"} << '\n';
         }
     }
-    // see what change we made to the outermost line number
-    if (update_line_only) {
-        const DILineInfo &frame = DI.at(nframes - 1 - nctx);
-        nctx += 1;
-        context.push_back(frame);
-        if (frame.Line != UINT_MAX && frame.Line != 0) {
-            StringRef method = StringRef(frame.FunctionName).rtrim(';');
-            Out << LineStart << inlining_indent("│")
-                << " @ " << frame.FileName
-                << ":" << frame.Line
-                << " within `" << method << "'";
-            if (collapse_recursive) {
-                while (nctx < nframes) {
-                    const DILineInfo &frame = DI.at(nframes - 1 - nctx);
-                    if (StringRef(frame.FunctionName).rtrim(';') != method)
-                        break;
-                    nctx += 1;
-                    context.push_back(frame);
-                    Out << " @ " << frame.FileName
-                        << ":" << frame.Line;
-                }
-            }
-            Out << "\n";
-        }
-    }
-    // now print the rest of the new frames
+    // print the new frames
     while (nctx < nframes) {
         const DILineInfo &frame = DI.at(nframes - 1 - nctx);
         Out << LineStart << inlining_indent("│");
         nctx += 1;
         context.push_back(frame);
-        this->inline_depth += 1;
-        if (bracket_outer || nctx != 1)
-            Out << "┌";
+        if (update_line_only) {
+            update_line_only = false;
+        }
+        else {
+            this->inline_depth += 1;
+            if (bracket_outer || nctx != 1)
+                Out << "┌";
+        }
         Out << " @ " << frame.FileName;
         if (frame.Line != UINT_MAX && frame.Line != 0)
             Out << ":" << frame.Line;
-        Out << " within `" << StringRef(frame.FunctionName).rtrim(';') << "'";
+        StringRef method = StringRef(frame.FunctionName).rtrim(';');
+        Out << " within `" << method << "'";
         if (collapse_recursive) {
-            StringRef method = StringRef(frame.FunctionName).rtrim(';');
             while (nctx < nframes) {
                 const DILineInfo &frame = DI.at(nframes - 1 - nctx);
                 if (StringRef(frame.FunctionName).rtrim(';') != method)
@@ -334,8 +317,8 @@ void LineNumberAnnotatedWriter::emitFunctionAnnot(
     if (FuncLoc) {
         std::vector<DILineInfo> DIvec(1);
         DILineInfo &DI = DIvec.back();
-        DI.FunctionName = FuncLoc->getName();
-        DI.FileName = FuncLoc->getFilename();
+        DI.FunctionName = FuncLoc->getName().str();
+        DI.FileName = FuncLoc->getFilename().str();
         DI.Line = FuncLoc->getLine();
         LinePrinter.emit_lineinfo(Out, DIvec);
     }
@@ -358,8 +341,8 @@ void LineNumberAnnotatedWriter::emitInstructionAnnot(
             DILineInfo &DI = DIvec.back();
             DIScope *scope = NewInstrLoc->getScope();
             if (scope)
-                DI.FunctionName = scope->getName();
-            DI.FileName = NewInstrLoc->getFilename();
+                DI.FunctionName = scope->getName().str();
+            DI.FileName = NewInstrLoc->getFilename().str();
             DI.Line = NewInstrLoc->getLine();
             NewInstrLoc = NewInstrLoc->getInlinedAt();
         } while (NewInstrLoc);
@@ -375,123 +358,145 @@ void LineNumberAnnotatedWriter::emitBasicBlockEndAnnot(
         LinePrinter.emit_finish(Out);
 }
 
+static void jl_strip_llvm_debug(Module *m, bool all_meta, LineNumberAnnotatedWriter *AAW)
+{
+    // strip metadata from all instructions in all functions in the module
+    Instruction *deletelast = nullptr; // can't actually delete until the iterator advances
+    for (Function &f : m->functions()) {
+        if (AAW)
+            AAW->addSubprogram(&f, f.getSubprogram());
+        for (BasicBlock &f_bb : f) {
+            for (Instruction &inst : f_bb) {
+                if (deletelast) {
+                    deletelast->eraseFromParent();
+                    deletelast = nullptr;
+                }
+                // remove dbg.declare and dbg.value calls
+                if (isa<DbgDeclareInst>(inst) || isa<DbgValueInst>(inst)) {
+                    deletelast = &inst;
+                    continue;
+                }
+
+                // iterate over all metadata kinds and set to NULL to remove
+                if (all_meta) {
+                    SmallVector<std::pair<unsigned, MDNode*>, 4> MDForInst;
+                    inst.getAllMetadataOtherThanDebugLoc(MDForInst);
+                    for (const auto &md_iter : MDForInst) {
+                        inst.setMetadata(md_iter.first, NULL);
+                    }
+                }
+                // record debug location before erasing it
+                if (AAW)
+                    AAW->addDebugLoc(&inst, inst.getDebugLoc());
+                inst.setDebugLoc(DebugLoc());
+            }
+            if (deletelast) {
+                deletelast->eraseFromParent();
+                deletelast = nullptr;
+            }
+        }
+        f.setSubprogram(NULL);
+    }
+    if (all_meta) {
+        for (GlobalObject &g : m->global_objects()) {
+            g.clearMetadata();
+        }
+    }
+    // now that the subprogram is not referenced, we can delete it too
+    if (NamedMDNode *md = m->getNamedMetadata("llvm.dbg.cu"))
+        m->eraseNamedMetadata(md);
+    //if (NamedMDNode *md = m->getNamedMetadata("llvm.module.flags"))
+    //    m->eraseNamedMetadata(md);
+}
+
+void jl_strip_llvm_debug(Module *m)
+{
+    jl_strip_llvm_debug(m, false, NULL);
+}
+
+void jl_strip_llvm_addrspaces(Module *m)
+{
+    legacy::PassManager PM;
+    PM.add(createRemoveJuliaAddrspacesPass());
+    PM.run(*m);
+}
+
 // print an llvm IR acquired from jl_get_llvmf
 // warning: this takes ownership of, and destroys, f->getParent()
 extern "C" JL_DLLEXPORT
-jl_value_t *jl_dump_function_ir(void *f, bool strip_ir_metadata, bool dump_module, const char *debuginfo)
+jl_value_t *jl_dump_function_ir(void *f, char strip_ir_metadata, char dump_module, const char *debuginfo)
 {
     std::string code;
-    llvm::raw_string_ostream stream(code);
+    raw_string_ostream stream(code);
 
-    Function *llvmf = dyn_cast_or_null<Function>((Function*)f);
-    if (!llvmf || (!llvmf->isDeclaration() && !llvmf->getParent()))
-        jl_error("jl_dump_function_ir: Expected Function* in a temporary Module");
+    {
+        Function *llvmf = dyn_cast_or_null<Function>((Function*)f);
+        if (!llvmf || (!llvmf->isDeclaration() && !llvmf->getParent()))
+            jl_error("jl_dump_function_ir: Expected Function* in a temporary Module");
 
-    JL_LOCK(&codegen_lock); // Might GC
-    LineNumberAnnotatedWriter AAW{debuginfo};
-    if (!llvmf->getParent()) {
-        // print the function declaration as-is
-        llvmf->print(stream, &AAW);
-        delete llvmf;
-    }
-    else {
-        Module *m = llvmf->getParent();
-        if (strip_ir_metadata) {
-            // strip metadata from all instructions in all functions in the module
-            Instruction *deletelast = nullptr; // can't actually delete until the iterator advances
-            for (Function &f2 : m->functions()) {
-                AAW.addSubprogram(&f2, f2.getSubprogram());
-                for (BasicBlock &f2_bb : f2) {
-                    for (Instruction &inst : f2_bb) {
-                        if (deletelast) {
-                            deletelast->eraseFromParent();
-                            deletelast = nullptr;
-                        }
-                        // remove dbg.declare and dbg.value calls
-                        if (isa<DbgDeclareInst>(inst) || isa<DbgValueInst>(inst)) {
-                            deletelast = &inst;
-                            continue;
-                        }
-
-                        // iterate over all metadata kinds and set to NULL to remove
-                        SmallVector<std::pair<unsigned, MDNode*>, 4> MDForInst;
-                        inst.getAllMetadataOtherThanDebugLoc(MDForInst);
-                        for (const auto &md_iter : MDForInst) {
-                            inst.setMetadata(md_iter.first, NULL);
-                        }
-                        // record debug location before erasing it
-                        AAW.addDebugLoc(&inst, inst.getDebugLoc());
-                        inst.setDebugLoc(DebugLoc());
-                    }
-                    if (deletelast) {
-                        deletelast->eraseFromParent();
-                        deletelast = nullptr;
-                    }
-                }
-            }
-            for (GlobalObject &g2 : m->global_objects()) {
-                g2.clearMetadata();
-            }
-        }
-        if (dump_module) {
-            m->print(stream, &AAW);
+        JL_LOCK(&codegen_lock); // Might GC
+        LineNumberAnnotatedWriter AAW{debuginfo};
+        if (!llvmf->getParent()) {
+            // print the function declaration as-is
+            llvmf->print(stream, &AAW);
+            delete llvmf;
         }
         else {
-            llvmf->print(stream, &AAW);
+            Module *m = llvmf->getParent();
+            if (strip_ir_metadata) {
+                std::string llvmfn(llvmf->getName());
+                jl_strip_llvm_addrspaces(m);
+                jl_strip_llvm_debug(m, true, &AAW);
+                // rewriting the function type creates a new function, so look it up again
+                llvmf = m->getFunction(llvmfn);
+            }
+            if (dump_module) {
+                m->print(stream, &AAW);
+            }
+            else {
+                llvmf->print(stream, &AAW);
+            }
+            delete m;
         }
-        delete m;
+        JL_UNLOCK(&codegen_lock); // Might GC
     }
-    JL_UNLOCK(&codegen_lock); // Might GC
 
     return jl_pchar_to_string(stream.str().data(), stream.str().size());
 }
 
 static void jl_dump_asm_internal(
         uintptr_t Fptr, size_t Fsize, int64_t slide,
-        const object::ObjectFile *object,
+        object::SectionRef Section,
         DIContext *di_ctx,
         raw_ostream &rstream,
         const char* asm_variant,
         const char* debuginfo);
 
 // This isn't particularly fast, but neither is printing assembly, and they're only used for interactive mode
-static uint64_t compute_obj_symsize(const object::ObjectFile *obj, uint64_t offset)
+static uint64_t compute_obj_symsize(object::SectionRef Section, uint64_t offset)
 {
-    // Scan the object file for the closest symbols above and below offset in the .text section
+    // Scan the object file for the closest symbols above and below offset in the given section
     uint64_t lo = 0;
     uint64_t hi = 0;
     bool setlo = false;
-    for (const object::SectionRef &Section : obj->sections()) {
-        uint64_t SAddr, SSize;
-        if (!Section.isText()) continue;
-        SAddr = Section.getAddress();
-        SSize = Section.getSize();
-        if (offset < SAddr || offset >= SAddr + SSize) continue;
-        assert(hi == 0);
-
-        // test for lower and upper symbol bounds relative to other symbols
-        hi = SAddr + SSize;
-        object::section_iterator ESection = obj->section_end();
-        for (const object::SymbolRef &Sym : obj->symbols()) {
-            uint64_t Addr;
-            object::section_iterator Sect = ESection;
-            auto SectOrError = Sym.getSection();
-            assert(SectOrError);
-            Sect = SectOrError.get();
-            if (Sect == ESection) continue;
-            if (Sect != Section) continue;
-            auto AddrOrError = Sym.getAddress();
-            assert(AddrOrError);
-            Addr = AddrOrError.get();
-            if (Addr <= offset && Addr >= lo) {
-                // test for lower bound on symbol
-                lo = Addr;
-                setlo = true;
-            }
-            if (Addr > offset && Addr < hi) {
-                // test for upper bound on symbol
-                hi = Addr;
-            }
+    uint64_t SAddr = Section.getAddress();
+    uint64_t SSize = Section.getSize();
+    if (offset < SAddr || offset >= SAddr + SSize)
+        return 0;
+    // test for lower and upper symbol bounds relative to other symbols
+    hi = SAddr + SSize;
+    for (const object::SymbolRef &Sym : Section.getObject()->symbols()) {
+        if (!Section.containsSymbol(Sym))
+            continue;
+        uint64_t Addr = cantFail(Sym.getAddress());
+        if (Addr <= offset && Addr >= lo) {
+            // test for lower bound on symbol
+            lo = Addr;
+            setlo = true;
+        }
+        if (Addr > offset && Addr < hi) {
+            // test for upper bound on symbol
+            hi = Addr;
         }
     }
     if (setlo)
@@ -506,22 +511,22 @@ jl_value_t *jl_dump_fptr_asm(uint64_t fptr, int raw_mc, const char* asm_variant,
     assert(fptr != 0);
     jl_ptls_t ptls = jl_get_ptls_states();
     std::string code;
-    llvm::raw_string_ostream stream(code);
+    raw_string_ostream stream(code);
 
     // Find debug info (line numbers) to print alongside
+    object::SectionRef Section;
+    int64_t slide = 0;
     uint64_t symsize = 0;
-    int64_t slide = 0, section_slide = 0;
     llvm::DIContext *context = NULL;
-    const object::ObjectFile *object = NULL;
-    if (!jl_DI_for_fptr(fptr, &symsize, &slide, &section_slide, &object, &context)) {
-        if (!jl_dylib_DI_for_fptr(fptr, &object, &context, &slide, &section_slide, false,
-            NULL, NULL, NULL, NULL)) {
-                jl_printf(JL_STDERR, "WARNING: Unable to find function pointer\n");
-                return jl_pchar_to_string("", 0);
+    if (!jl_DI_for_fptr(fptr, &symsize, &slide, &Section, &context)) {
+        if (!jl_dylib_DI_for_fptr(fptr, &Section, &slide, &context,
+                    false, NULL, NULL, NULL, NULL)) {
+            jl_printf(JL_STDERR, "WARNING: Unable to find function pointer\n");
+            return jl_pchar_to_string("", 0);
         }
     }
-    if (symsize == 0 && object != NULL)
-        symsize = compute_obj_symsize(object, fptr + slide + section_slide);
+    if (symsize == 0 && Section.getObject())
+        symsize = compute_obj_symsize(Section, fptr + slide);
     if (symsize == 0) {
         jl_printf(JL_STDERR, "WARNING: Could not determine size of symbol\n");
         return jl_pchar_to_string("", 0);
@@ -535,7 +540,7 @@ jl_value_t *jl_dump_fptr_asm(uint64_t fptr, int raw_mc, const char* asm_variant,
     int8_t gc_state = jl_gc_safe_enter(ptls);
     jl_dump_asm_internal(
             fptr, symsize, slide,
-            object, context,
+            Section, context,
             stream,
             asm_variant,
             debuginfo);
@@ -602,20 +607,16 @@ const char *SymbolTable::lookupLocalPC(size_t addr) {
 
 StringRef SymbolTable::getSymbolNameAt(uint64_t offset) const
 {
-    if (object == NULL) return StringRef();
+    if (object == NULL)
+        return StringRef();
     object::section_iterator ESection = object->section_end();
     for (const object::SymbolRef &Sym : object->symbols()) {
-        uint64_t Addr, SAddr;
-        object::section_iterator Sect = ESection;
-        auto SectOrError = Sym.getSection();
-        assert(SectOrError);
-        Sect = SectOrError.get();
-        if (Sect == ESection) continue;
-        SAddr = Sect->getAddress();
-        if (SAddr == 0) continue;
-        auto AddrOrError = Sym.getAddress();
-        assert(AddrOrError);
-        Addr = AddrOrError.get();
+        auto Sect = cantFail(Sym.getSection());
+        if (Sect == ESection)
+            continue;
+        if (Sect->getAddress() == 0)
+            continue;
+        uint64_t Addr = cantFail(Sym.getAddress());
         if (Addr == offset) {
             auto sNameOrError = Sym.getName();
             if (sNameOrError)
@@ -640,14 +641,15 @@ void SymbolTable::createSymbols()
         uintptr_t rel = isymb->first - ip;
         uintptr_t addr = isymb->first;
         if (Fptr <= addr && addr < Fptr + Fsize) {
-            std::ostringstream name;
-            name << "L" << rel;
-            isymb->second = name.str();
+            std::string name;
+            raw_string_ostream(name) << "L" << rel;
+            isymb->second = name;
         }
         else {
             const char *global = lookupLocalPC(addr);
-            if (global)
+            if (global && global[0])
                 isymb->second = global;
+            // TODO: free(global)?
         }
     }
 }
@@ -663,8 +665,8 @@ const char *SymbolTable::lookupSymbolName(uint64_t addr)
         if (local_name.empty()) {
             const char *global = lookupLocalPC(addr);
             if (global) {
-                //std::ostringstream name;
-                //name << global << "@0x" << std::hex
+                //std::string name;
+                //raw_string_ostream(name) << global << "@0x" << std::hex
                 //     << std::setfill('0') << std::setw(2 * sizeof(void*))
                 //     << addr;
                 //Sym->second = name.str();
@@ -672,7 +674,7 @@ const char *SymbolTable::lookupSymbolName(uint64_t addr)
             }
         }
         else {
-            Sym->second = local_name;
+            Sym->second = local_name.str();
         }
     }
     return Sym->second.empty() ? NULL : Sym->second.c_str();
@@ -740,7 +742,7 @@ static int OpInfoLookup(void *DisInfo, uint64_t PC, uint64_t Offset, uint64_t Si
 
 static void jl_dump_asm_internal(
         uintptr_t Fptr, size_t Fsize, int64_t slide,
-        const object::ObjectFile *object,
+        object::SectionRef Section,
         DIContext *di_ctx,
         raw_ostream &rstream,
         const char* asm_variant,
@@ -758,10 +760,14 @@ static void jl_dump_asm_internal(
     const Target *TheTarget = TargetRegistry::lookupTarget(TheTriple.str(), err);
 
     // Set up required helpers and streamer
-    std::unique_ptr<MCStreamer> Streamer;
     SourceMgr SrcMgr;
 
-    std::unique_ptr<MCAsmInfo> MAI(TheTarget->createMCAsmInfo(*TheTarget->createMCRegInfo(TheTriple.str()), TheTriple.str()));
+    MCTargetOptions Options;
+    std::unique_ptr<MCAsmInfo> MAI(TheTarget->createMCAsmInfo(*TheTarget->createMCRegInfo(TheTriple.str()), TheTriple.str()
+#if JL_LLVM_VERSION >= 100000
+            , Options
+#endif
+        ));
     assert(MAI && "Unable to create target asm info!");
 
     std::unique_ptr<MCRegisterInfo> MRI(TheTarget->createMCRegInfo(TheTriple.str()));
@@ -786,53 +792,40 @@ static void jl_dump_asm_internal(
     }
     bool ShowEncoding = false;
 
-    std::unique_ptr<MCInstrInfo> MCII(TheTarget->createMCInstrInfo());
-    std::unique_ptr<MCInstrAnalysis>
-        MCIA(TheTarget->createMCInstrAnalysis(MCII.get()));
-    MCInstPrinter *IP =
-        TheTarget->createMCInstPrinter(TheTriple, OutputAsmVariant, *MAI, *MCII, *MRI);
+    std::unique_ptr<MCInstrInfo> MCII(
+            TheTarget->createMCInstrInfo());
+    std::unique_ptr<MCInstrAnalysis> MCIA(
+            TheTarget->createMCInstrAnalysis(MCII.get()));
+    std::unique_ptr<MCInstPrinter> IP(
+            TheTarget->createMCInstPrinter(TheTriple, OutputAsmVariant, *MAI, *MCII, *MRI));
     //IP->setPrintImmHex(true); // prefer hex or decimal immediates
-#if JL_LLVM_VERSION >= 70000
-    std::unique_ptr<MCCodeEmitter> CE = 0;
-    std::unique_ptr<MCAsmBackend> MAB = 0;
+    std::unique_ptr<MCCodeEmitter> CE;
+    std::unique_ptr<MCAsmBackend> MAB;
     if (ShowEncoding) {
-        CE = std::unique_ptr<MCCodeEmitter>(TheTarget->createMCCodeEmitter(*MCII, *MRI, Ctx));
-        MCTargetOptions Options;
-        MAB = std::unique_ptr<MCAsmBackend>(TheTarget->createMCAsmBackend(*STI, *MRI, Options));
+        CE.reset(TheTarget->createMCCodeEmitter(*MCII, *MRI, Ctx));
+        MAB.reset(TheTarget->createMCAsmBackend(*STI, *MRI, Options));
     }
-#else
-    MCCodeEmitter* CE = 0;
-    MCAsmBackend* MAB = 0;
-    if (ShowEncoding) {
-        CE = TheTarget->createMCCodeEmitter(*MCII, *MRI, Ctx);
-        MCTargetOptions Options;
-        MAB = TheTarget->createMCAsmBackend(*STI, *MRI, Options);
-    }
-#endif
 
     // createAsmStreamer expects a unique_ptr to a formatted stream, which means
     // it will destruct the stream when it is done. We cannot have this, so we
     // start out with a raw stream, and create formatted stream from it here.
-    // LLVM will desctruct the formatted stream, and we keep the raw stream.
-    auto ustream = llvm::make_unique<formatted_raw_ostream>(rstream);
-#if JL_LLVM_VERSION >= 70000
-    Streamer.reset(TheTarget->createAsmStreamer(Ctx, std::move(ustream), /*asmverbose*/true,
-                                                /*useDwarfDirectory*/ true,
-                                                IP, std::move(CE), std::move(MAB), /*ShowInst*/ false));
-#else
-    Streamer.reset(TheTarget->createAsmStreamer(Ctx, std::move(ustream), /*asmverbose*/true,
-                                                /*useDwarfDirectory*/ true,
-                                                IP, CE, MAB, /*ShowInst*/ false));
-#endif
+    // LLVM will destroy the formatted stream, and we keep the raw stream.
+    std::unique_ptr<formatted_raw_ostream> ustream(new formatted_raw_ostream(rstream));
+    std::unique_ptr<MCStreamer> Streamer(
+            TheTarget->createAsmStreamer(Ctx, std::move(ustream), /*asmverbose*/true,
+                                         /*useDwarfDirectory*/ true,
+                                         IP.release(),
+                                         std::move(CE), std::move(MAB),
+                                         /*ShowInst*/ false));
     Streamer->InitSections(true);
 
     // Make the MemoryObject wrapper
     ArrayRef<uint8_t> memoryObject(const_cast<uint8_t*>((const uint8_t*)Fptr),Fsize);
-    SymbolTable DisInfo(Ctx, object, slide, memoryObject);
+    SymbolTable DisInfo(Ctx, Section.getObject(), slide, memoryObject);
 
     DILineInfoTable di_lineinfo;
     if (di_ctx)
-        di_lineinfo = di_ctx->getLineInfoForAddressRange(makeAddress(Fptr + slide), Fsize);
+        di_lineinfo = di_ctx->getLineInfoForAddressRange(makeAddress(Section, Fptr + slide), Fsize);
     if (!di_lineinfo.empty()) {
         auto cur_addr = di_lineinfo[0].first;
         auto nlineinfo = di_lineinfo.size();
@@ -885,7 +878,11 @@ static void jl_dump_asm_internal(
                     std::string buf;
                     dbgctx.emit_lineinfo(buf, di_lineIter->second);
                     if (!buf.empty()) {
+#if JL_LLVM_VERSION >= 110000
+                        Streamer->emitRawText(buf);
+#else
                         Streamer->EmitRawText(buf);
+#endif
                     }
                 }
             }
@@ -900,17 +897,27 @@ static void jl_dump_asm_internal(
             if (pass != 0 && nextLineAddr != (uint64_t)-1 && Index + Fptr + slide == nextLineAddr) {
                 if (di_ctx) {
                     std::string buf;
-                    DILineInfoSpecifier infoSpec(DILineInfoSpecifier::FileLineInfoKind::Default,
-                                                 DILineInfoSpecifier::FunctionNameKind::ShortName);
-                    DIInliningInfo dbg = di_ctx->getInliningInfoForAddress(makeAddress(Index + Fptr + slide), infoSpec);
+                    DILineInfoSpecifier infoSpec(
+#if JL_LLVM_VERSION >= 110000
+                        DILineInfoSpecifier::FileLineInfoKind::RawValue,
+#else
+                        DILineInfoSpecifier::FileLineInfoKind::Default,
+#endif
+                        DILineInfoSpecifier::FunctionNameKind::ShortName);
+                    DIInliningInfo dbg = di_ctx->getInliningInfoForAddress(makeAddress(Section, Index + Fptr + slide), infoSpec);
                     if (dbg.getNumberOfFrames()) {
                         dbgctx.emit_lineinfo(buf, dbg);
                     }
                     else {
                         dbgctx.emit_lineinfo(buf, di_lineIter->second);
                     }
-                    if (!buf.empty())
+                    if (!buf.empty()) {
+#if JL_LLVM_VERSION >= 110000
+                        Streamer->emitRawText(buf);
+#else
                         Streamer->EmitRawText(buf);
+#endif
+                    }
                     nextLineAddr = (++di_lineIter)->first;
                 }
             }
@@ -920,15 +927,22 @@ static void jl_dump_asm_internal(
                 // Uncomment this to output addresses for all instructions
                 // stream << Index << ": ";
                 MCSymbol *symbol = DisInfo.lookupSymbol(Fptr+Index);
-                if (symbol)
+                if (symbol) {
+#if JL_LLVM_VERSION >= 110000
+                    Streamer->emitLabel(symbol);
+#else
                     Streamer->EmitLabel(symbol);
+#endif
+                }
             }
 
             MCInst Inst;
             MCDisassembler::DecodeStatus S;
             FuncMCView view = memoryObject.slice(Index);
             S = DisAsm->getInstruction(Inst, insSize, view, 0,
+#if JL_LLVM_VERSION < 100000
                                       /*VStream*/ nulls(),
+#endif
                                       /*CStream*/ pass != 0 ? Streamer->GetCommentOS() : nulls());
             if (pass != 0 && Streamer->GetCommentOS().tell() > 0)
                 Streamer->GetCommentOS() << '\n';
@@ -941,23 +955,34 @@ static void jl_dump_asm_internal(
                     insSize = 1; // attempt to slide 1 byte forward
 #endif
                 if (pass != 0) {
-                    std::ostringstream buf;
-                    if (insSize == 4)
-                        buf << "\t.long\t0x" << std::hex
-                            << std::setfill('0') << std::setw(8)
-                            << *(uint32_t*)(Fptr+Index);
-                    else
-                        for (uint64_t i=0; i<insSize; ++i)
-                            buf << "\t.byte\t0x" << std::hex
-                                << std::setfill('0') << std::setw(2)
-                                << (int)*(uint8_t*)(Fptr+Index+i);
+                    std::string _buf;
+                    raw_string_ostream buf(_buf);
+                    if (insSize == 4) {
+                        buf << "\t.long\t";
+                        llvm::write_hex(buf, *(uint32_t*)(Fptr + Index), HexPrintStyle::PrefixLower, 8);
+                    }
+                    else {
+                        for (uint64_t i = 0; i < insSize; ++i) {
+                            buf << "\t.byte\t";
+                            llvm::write_hex(buf, *(uint8_t*)(Fptr + Index + i), HexPrintStyle::PrefixLower, 2);
+                        }
+                    }
+#if JL_LLVM_VERSION >= 110000
+                    Streamer->emitRawText(StringRef(buf.str()));
+#else
                     Streamer->EmitRawText(StringRef(buf.str()));
+#endif
                 }
                 break;
 
             case MCDisassembler::SoftFail:
-                if (pass != 0)
+                if (pass != 0) {
+#if JL_LLVM_VERSION >= 110000
+                    Streamer->emitRawText(StringRef("potentially undefined instruction encoding:"));
+#else
                     Streamer->EmitRawText(StringRef("potentially undefined instruction encoding:"));
+#endif
+                }
                 // Fall through
 
             case MCDisassembler::Success:
@@ -989,7 +1014,11 @@ static void jl_dump_asm_internal(
                             }
                         }
                     }
+#if JL_LLVM_VERSION >= 110000
+                    Streamer->emitInstruction(Inst, *STI);
+#else
                     Streamer->EmitInstruction(Inst, *STI);
+#endif
                 }
                 break;
             }
@@ -1003,7 +1032,11 @@ static void jl_dump_asm_internal(
             std::string buf;
             dbgctx.emit_finish(buf);
             if (!buf.empty()) {
+#if JL_LLVM_VERSION >= 110000
+                Streamer->emitRawText(buf);
+#else
                 Streamer->EmitRawText(buf);
+#endif
             }
         }
     }
