@@ -7,27 +7,25 @@
 #include "llvm/IR/LegacyPassManager.h"
 
 #include <llvm/ExecutionEngine/SectionMemoryManager.h>
-#include "llvm/ExecutionEngine/Orc/CompileUtils.h"
-#include "llvm/ExecutionEngine/Orc/IRCompileLayer.h"
-#include "llvm/ExecutionEngine/Orc/LambdaResolver.h"
-#include "llvm/ExecutionEngine/Orc/LazyEmittingLayer.h"
-#include "llvm/ExecutionEngine/Orc/RTDyldObjectLinkingLayer.h"
-#include "llvm/ExecutionEngine/JITEventListener.h"
+#include <llvm/ExecutionEngine/Orc/IRCompileLayer.h>
+#include <llvm/ExecutionEngine/Orc/RTDyldObjectLinkingLayer.h>
+#include <llvm/ExecutionEngine/JITEventListener.h>
 
 #include <llvm/Target/TargetMachine.h>
 #include "julia_assert.h"
 
-extern "C" {
-    extern int globalUnique;
-}
+using namespace llvm;
+
 extern TargetMachine *jl_TargetMachine;
-extern Module *shadow_output;
 extern bool imaging_mode;
 
 void addTargetPasses(legacy::PassManagerBase *PM, TargetMachine *TM);
 void addOptimizationPasses(legacy::PassManagerBase *PM, int opt_level, bool lower_intrinsics=true, bool dump_native=false);
+void addMachinePasses(legacy::PassManagerBase *PM, TargetMachine *TM);
 void jl_finalize_module(std::unique_ptr<Module>  m);
 void jl_merge_module(Module *dest, std::unique_ptr<Module> src);
+Module *jl_create_llvm_module(StringRef name);
+GlobalVariable *jl_emit_RTLD_DEFAULT_var(Module *M);
 
 typedef struct _jl_llvm_functions_t {
     std::string functionObject;     // jlcall llvm Function name
@@ -53,11 +51,36 @@ typedef std::vector<std::tuple<jl_code_instance_t*, jl_returninfo_t::CallingConv
 typedef std::tuple<std::unique_ptr<Module>, jl_llvm_functions_t> jl_compile_result_t;
 
 typedef struct {
+    typedef StringMap<GlobalVariable*> SymMapGV;
     // outputs
     jl_codegen_call_targets_t workqueue;
     std::map<void*, GlobalVariable*> globals;
     std::map<jl_datatype_t*, DIType*> ditypes;
     std::map<jl_datatype_t*, Type*> llvmtypes;
+    DenseMap<Constant*, GlobalVariable*> mergedConstants;
+    // Map from symbol name (in a certain library) to its GV in sysimg and the
+    // DL handle address in the current session.
+    StringMap<std::pair<GlobalVariable*,SymMapGV>> libMapGV;
+#ifdef _OS_WINDOWS_
+    SymMapGV symMapExe;
+    SymMapGV symMapDl;
+#endif
+    SymMapGV symMapDefault;
+    // Map from distinct callee's to its GOT entry.
+    // In principle the attribute, function type and calling convention
+    // don't need to be part of the key but it seems impossible to forward
+    // all the arguments without writing assembly directly.
+    // This doesn't matter too much in reality since a single function is usually
+    // not called with multiple signatures.
+    DenseMap<AttributeList, std::map<
+        std::tuple<GlobalVariable*, FunctionType*, CallingConv::ID>,
+        GlobalVariable*>> allPltMap;
+    Module *_shared_module = NULL;
+    Module *shared_module(LLVMContext &context) {
+        if (!_shared_module)
+            _shared_module = jl_create_llvm_module("globals");
+        return _shared_module;
+    }
     // inputs
     size_t world = 0;
     const jl_cgparams_t *params = &jl_default_cgparams;
@@ -88,41 +111,7 @@ void jl_compile_workqueue(
 Function *jl_cfunction_object(jl_function_t *f, jl_value_t *rt, jl_tupletype_t *argt,
     jl_codegen_params_t &params);
 
-// Connect Modules via prototypes, each owned by module `M`
-static inline GlobalVariable *global_proto(GlobalVariable *G, Module *M = NULL)
-{
-    // Copy the GlobalVariable, but without the initializer, so it becomes a declaration
-    GlobalVariable *proto = new GlobalVariable(G->getType()->getElementType(),
-            G->isConstant(), GlobalVariable::ExternalLinkage,
-            NULL, G->getName(),  G->getThreadLocalMode());
-    proto->copyAttributesFrom(G);
-    // DLLImport only needs to be set for the shadow module
-    // it just gets annoying in the JIT
-    proto->setDLLStorageClass(GlobalValue::DefaultStorageClass);
-    if (M)
-        M->getGlobalList().push_back(proto);
-    return proto;
-}
-
-static inline GlobalVariable *prepare_global_in(Module *M, GlobalVariable *G)
-{
-    if (G->getParent() == M)
-        return G;
-    GlobalValue *local = M->getNamedValue(G->getName());
-    if (!local) {
-        local = global_proto(G, M);
-    }
-    return cast<GlobalVariable>(local);
-}
-
-void add_named_global(GlobalObject *gv, void *addr, bool dllimport);
-template<typename T>
-static inline void add_named_global(GlobalObject *gv, T *addr, bool dllimport = true)
-{
-    // cast through integer to avoid c++ pedantic warning about casting between
-    // data and code pointers
-    add_named_global(gv, (void*)(uintptr_t)addr, dllimport);
-}
+void add_named_global(StringRef name, void *addr);
 
 static inline Constant *literal_static_pointer_val(const void *p, Type *T)
 {
@@ -150,64 +139,65 @@ typedef JITSymbol JL_JITSymbol;
 // is expected.
 typedef JITSymbol JL_SymbolInfo;
 
+#if JL_LLVM_VERSION < 120000
 using RTDyldObjHandleT = orc::VModuleKey;
+#endif
+
 #if JL_LLVM_VERSION >= 100000
-using CompilerResultT = Expected<orc::LegacyRTDyldObjectLinkingLayerBase::ObjectPtr>;
+using CompilerResultT = Expected<std::unique_ptr<llvm::MemoryBuffer>>;
 #else
 using CompilerResultT = std::unique_ptr<llvm::MemoryBuffer>;
 #endif
 
 class JuliaOJIT {
-    // Custom object emission notification handler for the JuliaOJIT
-    class DebugObjectRegistrar {
-    public:
-        DebugObjectRegistrar(JuliaOJIT &JIT);
-        template <typename ObjSetT, typename LoadResult>
-        void operator()(RTDyldObjHandleT H, const ObjSetT &Object, const LoadResult &LOS);
-    private:
-        template <typename ObjT, typename LoadResult>
-        void registerObject(RTDyldObjHandleT H, const ObjT &Obj, const LoadResult &LO);
-        std::unique_ptr<JITEventListener> JuliaListener;
-        JuliaOJIT &JIT;
-    };
-
-    struct CompilerT {
+    struct CompilerT : public orc::IRCompileLayer::IRCompiler {
         CompilerT(JuliaOJIT *pjit)
-            : jit(*pjit)
-        {}
-        CompilerResultT operator()(Module &M);
+            : IRCompiler(orc::IRSymbolMapper::ManglingOptions{}),
+              jit(*pjit) {}
+        virtual CompilerResultT operator()(Module &M) override;
     private:
         JuliaOJIT &jit;
     };
+#if JL_LLVM_VERSION >= 120000
+    // Custom object emission notification handler for the JuliaOJIT
+    template <typename ObjT, typename LoadResult>
+    void registerObject(const ObjT &Obj, const LoadResult &LO);
+#else
+    // Custom object emission notification handler for the JuliaOJIT
+    template <typename ObjT, typename LoadResult>
+    void registerObject(RTDyldObjHandleT H, const ObjT &Obj, const LoadResult &LO);
+#endif
 
 public:
-    typedef orc::LegacyRTDyldObjectLinkingLayer ObjLayerT;
-    typedef orc::LegacyIRCompileLayer<ObjLayerT,CompilerT> CompileLayerT;
-    typedef orc::VModuleKey ModuleHandleT;
-    typedef StringMap<void*> SymbolTableT;
+    typedef orc::RTDyldObjectLinkingLayer ObjLayerT;
+    typedef orc::IRCompileLayer CompileLayerT;
+#if JL_LLVM_VERSION < 120000
+    typedef RTDyldObjHandleT ModuleHandleT;
+#endif
     typedef object::OwningBinary<object::ObjectFile> OwningObj;
 
-    JuliaOJIT(TargetMachine &TM);
+    JuliaOJIT(TargetMachine &TM, LLVMContext *Ctx);
 
     void RegisterJITEventListener(JITEventListener *L);
+#if JL_LLVM_VERSION < 120000
     std::vector<JITEventListener *> EventListeners;
     void NotifyFinalizer(RTDyldObjHandleT Key,
                          const object::ObjectFile &Obj,
                          const RuntimeDyld::LoadedObjectInfo &LoadedObjectInfo);
+#endif
     void addGlobalMapping(StringRef Name, uint64_t Addr);
-    void addGlobalMapping(const GlobalValue *GV, void *Addr);
-    void *getPointerToGlobalIfAvailable(StringRef S);
-    void *getPointerToGlobalIfAvailable(const GlobalValue *GV);
     void addModule(std::unique_ptr<Module> M);
+#if JL_LLVM_VERSION < 120000
     void removeModule(ModuleHandleT H);
+#endif
     JL_JITSymbol findSymbol(StringRef Name, bool ExportedSymbolsOnly);
     JL_JITSymbol findUnmangledSymbol(StringRef Name);
-    JL_JITSymbol resolveSymbol(StringRef Name);
     uint64_t getGlobalValueAddress(StringRef Name);
     uint64_t getFunctionAddress(StringRef Name);
     StringRef getFunctionAtAddress(uint64_t Addr, jl_code_instance_t *codeinst);
     const DataLayout& getDataLayout() const;
     const Triple& getTargetTriple() const;
+    size_t getTotalBytes() const;
 private:
     std::string getMangledName(StringRef Name);
     std::string getMangledName(const GlobalValue *GV);
@@ -225,16 +215,17 @@ private:
     TargetMachine *TMs[4];
     MCContext *Ctx;
     std::shared_ptr<RTDyldMemoryManager> MemMgr;
-    DebugObjectRegistrar registrar;
+    std::unique_ptr<JITEventListener> JuliaListener;
 
-    llvm::orc::ExecutionSession ES;
-    std::shared_ptr<llvm::orc::SymbolResolver> SymbolResolver;
+
+    orc::ThreadSafeContext TSCtx;
+    orc::ExecutionSession ES;
+    orc::JITDylib &GlobalJD;
+    orc::JITDylib &JD;
 
     ObjLayerT ObjectLayer;
     CompileLayerT CompileLayer;
 
-    SymbolTableT GlobalSymbolTable;
-    SymbolTableT LocalSymbolTable;
     DenseMap<void*, StringRef> ReverseLocalSymbolTable;
 };
 extern JuliaOJIT *jl_ExecutionEngine;
@@ -247,8 +238,11 @@ Pass *createLowerExcHandlersPass();
 Pass *createGCInvariantVerifierPass(bool Strong);
 Pass *createPropagateJuliaAddrspaces();
 Pass *createRemoveJuliaAddrspacesPass();
+Pass *createRemoveNIPass();
+Pass *createJuliaLICMPass();
 Pass *createMultiVersioningPass();
 Pass *createAllocOptPass();
+Pass *createDemoteFloat16Pass();
 // Whether the Function is an llvm or julia intrinsic.
 static inline bool isIntrinsicFunction(Function *F)
 {
